from typing import Generic, List, TypeVar

import requests

from understand_sdk.const import DEFAULT_HOST
from understand_sdk.model import BaseModel
from understand_sdk.story import Story, StoryWithChannels
from understand_sdk.task import Event, Task


class CreatedStory(BaseModel):
    id: str


PayloadT = TypeVar("PayloadT")


class Response(BaseModel, Generic[PayloadT]):
    status: str
    payload: PayloadT


class UnderstandPublicClient:
    """ """

    def __init__(self, api_key: str, host: str = DEFAULT_HOST):
        self.host = host

        self.session = requests.Session()
        self.session.headers.update({"X-UL-API-KEY": api_key})

    def create_story(self, story: StoryWithChannels) -> CreatedStory:
        res = self.session.post(
            f"{self.host}/api/v1/public/stories", json=story.model_dump(exclude_none=True, by_alias=True)
        )
        res.raise_for_status()

        raw_data = res.json()
        data = Response[CreatedStory](**raw_data)
        return data.payload


class UnderstandClient:
    """ """

    def __init__(self, token: str, host: str = DEFAULT_HOST):
        self.host = host

        self.session = requests.Session()
        self.session.headers.update({"Authorization": f"Bearer {token}"})

    def get_task(self) -> Task:
        res = self.session.get(f"{self.host}/api/v1/internal/task")
        res.raise_for_status()

        raw_data = res.json()
        data = Response[Task].model_validate(raw_data)
        return data.payload

    def create_stories(self, stories: List[Story]):
        res = self.session.post(
            f"{self.host}/api/v1/internal/task/stories",
            json=[story.model_dump(exclude_none=True, by_alias=True) for story in stories],
        )
        res.raise_for_status()

    def create_task_events(self, events: List[Event]):
        self.session.post(
            f"{self.host}/api/v1/internal/task/events",
            json=[event.model_dump(exclude_none=True, by_alias=True) for event in events],
        )



from abc import ABC
from enum import Enum
from typing import Dict, List, Literal, Optional, Union

from pydantic import Field

from understand_sdk.model import BaseModel

#
#
# Story
#
#


class SlideLayout(str, Enum):
    ROW = "row"
    COLUMN = "column"


class Slide(BaseModel, ABC):
    title: Optional[str] = None


class SlideWithLayout(Slide, ABC):
    layout: Optional[SlideLayout] = None


class ChartSlide(SlideWithLayout, ABC):
    description: Optional[str] = None
    data: List[Dict[str, Union[str, float, int]]]


#
# Chart slide types attributes
#


class Attribute(BaseModel, ABC):
    field: str


class ElementAttribute(Attribute):
    role: Literal["element"] = "element"


class XAttribute(Attribute):
    role: Literal["x"] = "x"


class YAttribute(Attribute):
    role: Literal["y"] = "y"


class SizeByAttribute(Attribute):
    role: Literal["sizeBy"] = "sizeBy"


class TargetAttribute(Attribute):
    role: Literal["target"] = "target"


class VarianceAttribute(Attribute):
    role: Literal["variance"] = "variance"


class SeriesAttribute(Attribute):
    role: Literal["series"] = "series"
    from_: str = Field(..., alias="from")
    to: str


#
# Settings options
#


class Variance(str, Enum):
    ABSOLUTE = "absolute"
    RELATIVE = "relative"
    CUSTOM_BAR = "customBar"
    CUSTOM_DIVERGING_BARS = "customDivergingBars"


class TargetFillArea(str, Enum):
    ABOVE = "above"
    BELOW = "below"
    BOTH = "both"


#
# Text slides
#


class CoverSlide(Slide):
    type: Literal["cover"] = "cover"
    title: Optional[str] = None


class HeadingSlide(Slide):
    type: Literal["heading"] = "heading"
    title: Optional[str] = None


class ParagraphSlide(SlideWithLayout):
    type: Literal["paragraph"] = "paragraph"
    description: Optional[str] = None


#
# KPI slides
#


class KpiArrow(str, Enum):
    UP = "up"
    DOWN = "down"


class KpiSentiment(str, Enum):
    POSITIVE = "positive"
    NEUTRAL = "neutral"
    NEGATIVE = "negative"


class KpiChange(BaseModel):
    value: Optional[str] = None
    suffix: Optional[str] = Field(default=None, alias="valueSufix")  # TODO(vojta) fix typo on API
    arrow: KpiArrow
    sentiment: KpiSentiment


class KpiTarget(BaseModel):
    value: Optional[str] = None
    percentage: Optional[float] = None


class KpiInstance(BaseModel):
    measure: Optional[str] = None
    period: Optional[str] = None
    value: Optional[str] = None
    suffix: Optional[str] = Field(default=None, alias="valueSufix")  # TODO(vojta) fix typo on API
    change: Optional[KpiChange] = None
    target: Optional[KpiTarget] = None


class KpiSlide(SlideWithLayout):
    type: Literal["kpi"] = "kpi"
    instances: List[KpiInstance]


#
# Line slide
#


class LineSlideDisplay(str, Enum):
    BASIC = "basic"
    CHANGE = "change"
    TARGET = "target"


class LineSlideSettings(BaseModel):
    variance: Optional[Variance] = None
    display: Optional[LineSlideDisplay] = LineSlideDisplay.BASIC
    targetFillArea: Optional[TargetFillArea] = None


class LineSlide(ChartSlide):
    type: Literal["line"] = "line"
    settings: Optional[LineSlideSettings] = None
    attributes: List[
        Union[
            ElementAttribute,
            YAttribute,
            TargetAttribute,
            VarianceAttribute,
        ]
    ]


#
# Bar slide
#


class BarSlideDisplay(str, Enum):
    BASIC = "basic"
    CHANGE = "change"
    TARGET = "target"


class BarSlideSettings(BaseModel):
    variance: Optional[Variance] = None
    display: Optional[BarSlideDisplay] = BarSlideDisplay.BASIC


class BarSlide(ChartSlide):
    type: Literal["bar"] = "bar"
    settings: Optional[BarSlideSettings] = None
    attributes: List[
        Union[
            ElementAttribute,
            XAttribute,
            SeriesAttribute,
            VarianceAttribute,
        ]
    ]


#
# Column slide
#


class ColumnSlideDisplay(str, Enum):
    BASIC = "basic"
    CHANGE = "change"
    TARGET = "target"


class ColumnSlideSettings(BaseModel):
    variance: Optional[Variance] = None
    display: Optional[ColumnSlideDisplay] = ColumnSlideDisplay.BASIC


class ColumnSlide(ChartSlide):
    type: Literal["column"] = "column"
    settings: Optional[ColumnSlideSettings] = None
    attributes: List[
        Union[
            ElementAttribute,
            YAttribute,
            SeriesAttribute,
            VarianceAttribute,
        ]
    ]


#
# Dot slide
#


class DotSlideDisplay(str, Enum):
    BASIC = "basic"
    CHANGE = "change"


class DotSlideSettings(BaseModel):
    variance: Optional[Variance] = None
    display: Optional[DotSlideDisplay] = DotSlideDisplay.BASIC


class DotSlide(ChartSlide):
    type: Literal["dot"] = "dot"
    settings: Optional[DotSlideSettings] = None
    attributes: List[
        Union[
            ElementAttribute,
            YAttribute,
            SeriesAttribute,
            VarianceAttribute,
        ]
    ]


#
# Scatter slide
#


class ScatterSlide(ChartSlide):
    type: Literal["scatter"] = "scatter"
    attributes: List[Union[ElementAttribute, YAttribute, XAttribute, SeriesAttribute]]


#
# Bubble slide
#


class BubbleSlide(ChartSlide):
    type: Literal["bubble"] = "bubble"
    attributes: List[Union[ElementAttribute, YAttribute, XAttribute, SeriesAttribute, SizeByAttribute]]


#
# Story
#


class Story(BaseModel):
    title: Optional[str] = None
    slides: List[
        Union[
            CoverSlide,
            HeadingSlide,
            ParagraphSlide,
            KpiSlide,
            BubbleSlide,
            ScatterSlide,
            ColumnSlide,
            BarSlide,
            DotSlide,
            LineSlide,
        ]
    ]


class StoryWithChannels(Story):
    channels: Optional[List[str]] = None



from typing import Any, Dict, Optional

from pydantic import Field

from understand_sdk.model import BaseModel


class ConnectionConfig(BaseModel):
    version: int
    data: Dict[str, Any]


class Connection(BaseModel):
    type: str
    provider: str
    config: Optional[ConnectionConfig] = None


class Workspace(BaseModel):
    id: str
    slug: str


class Channel(BaseModel):
    id: str
    slug: str


class Task(BaseModel):
    id: str
    created_at: str = Field(alias="createdAt")
    updated_at: str = Field(alias="updatedAt")
    workspace: Workspace
    channel: Channel
    connection: Connection


class Event(BaseModel):
    group: str = "container"
    name: str
    value: str



from datetime import datetime
from typing import List

from understand_sdk.story import (
    BarSlide,
    BarSlideDisplay,
    BarSlideSettings,
    BubbleSlide,
    ColumnSlide,
    ColumnSlideDisplay,
    ColumnSlideSettings,
    CoverSlide,
    DotSlide,
    ElementAttribute,
    HeadingSlide,
    KpiArrow,
    KpiChange,
    KpiInstance,
    KpiSentiment,
    KpiSlide,
    KpiTarget,
    LineSlide,
    LineSlideDisplay,
    LineSlideSettings,
    ParagraphSlide,
    ScatterSlide,
    SeriesAttribute,
    SizeByAttribute,
    Slide,
    SlideLayout,
    Story,
    TargetAttribute,
    TargetFillArea,
    Variance,
    VarianceAttribute,
    XAttribute,
    YAttribute,
)


def create_story(dt: datetime) -> Story:
    return Story(
        title=f"Example story - {str(dt.date())}",
        slides=[
            # Cover
            CoverSlide(title=f"Example story\n{str(dt.date())}"),
            # Slide variant examples ...
            *_story_paragraph_slides(),
            *_story_kpi_slides(),
            *_story_scatter_slides(),
            *_story_bubble_slides(),
            *_story_dot_slides(),
            *_story_line_slides(),
            *_story_column_slides(),
            *_story_bar_slides(),
        ],
    )


def _story_paragraph_slides() -> List[Slide]:
    return [
        HeadingSlide(title="Paragraph examples"),
        ParagraphSlide(
            title="Custom paragraph default layout - row", description="super long\ntext with **bold** \n # bold baby"
        ),
        ParagraphSlide(
            layout=SlideLayout.COLUMN,
            title="Custom paragraph column layout",
            description="super long\ntext with **bold** \n # bold baby",
        ),
    ]


def _story_kpi_slides() -> List[Slide]:
    return [
        HeadingSlide(title="KPI examples"),
        KpiSlide(
            title="KPI row",
            instances=[
                KpiInstance(
                    measure="Metric",
                    period="Year",
                    value="$99k",
                    suffix="m",
                ),
                KpiInstance(
                    measure="Metric",
                    period="Year",
                    value="$99k",
                    suffix="m",
                    change=KpiChange(
                        value="$255k",
                        arrow=KpiArrow.UP,
                        sentiment=KpiSentiment.NEGATIVE,
                    ),
                ),
            ],
        ),
        KpiSlide(
            layout=SlideLayout.COLUMN,
            title="KPI column & with target",
            instances=[
                KpiInstance(
                    measure="Metric",
                    period="Year",
                    value="$99k",
                    valueSufix="m",
                ),
                KpiInstance(
                    measure="Metric",
                    period="Year",
                    value="$99k",
                    suffix="m",
                    change=KpiChange(
                        value="$255k",
                        arrow=KpiArrow.UP,
                        sentiment=KpiSentiment.NEGATIVE,
                    ),
                    target=KpiTarget(value="200", percentage=59),
                ),
            ],
        ),
    ]


def _story_scatter_slides() -> List[Slide]:
    return [
        HeadingSlide(title="Scatter examples"),
        ScatterSlide(
            title="Scatter",
            attributes=[
                ElementAttribute(field="Company name"),
                XAttribute(field="Employees"),
                YAttribute(field="Amount"),
            ],
            data=EXAMPLE_DATA_SCATTER,
        ),
        ScatterSlide(
            title="Scatter with series",
            attributes=[
                ElementAttribute(field="Company name"),
                XAttribute(field="Employees"),
                YAttribute(field="Amount"),
                SeriesAttribute(field="Quarter", from_="Q1", to="Q2"),
            ],
            data=EXAMPLE_DATA_SCATTER,
        ),
    ]


def _story_bubble_slides() -> List[Slide]:
    return [
        HeadingSlide(title="Bubble examples"),
        BubbleSlide(
            title="Bubble",
            attributes=[
                ElementAttribute(field="Company name"),
                XAttribute(field="Employees"),
                SizeByAttribute(field="Size by metric"),
                YAttribute(field="Amount"),
            ],
            data=EXAMPLE_DATA_BUBBLE,
        ),
        BubbleSlide(
            title="Bubble with series",
            attributes=[
                ElementAttribute(field="Company name"),
                XAttribute(field="Employees"),
                YAttribute(field="Amount"),
                SizeByAttribute(field="Size by metric"),
                SeriesAttribute(field="Quarter", from_="Q1", to="Q2"),
            ],
            data=EXAMPLE_DATA_BUBBLE,
        ),
    ]


def _story_dot_slides() -> List[Slide]:
    return [
        HeadingSlide(title="Dot examples"),
        DotSlide(
            title="Dot",
            attributes=[
                ElementAttribute(field="Company name"),
                YAttribute(field="Amount"),
            ],
            data=EXAMPLE_DATA_GENERIC,
        ),
        DotSlide(
            title="Dot with series",
            attributes=[
                ElementAttribute(field="Company name"),
                YAttribute(field="Amount"),
                SeriesAttribute(field="Quarter", from_="Q1", to="Q2"),
            ],
            data=EXAMPLE_DATA_GENERIC,
        ),
    ]


def _story_line_slides() -> List[Slide]:
    return [
        HeadingSlide(title="Line examples"),
        LineSlide(
            title="Line",
            attributes=[
                ElementAttribute(field="Company name"),
                YAttribute(field="Amount"),
            ],
            data=EXAMPLE_DATA_LINE,
        ),
        LineSlide(
            title="Line with target",
            settings=LineSlideSettings(display=LineSlideDisplay.TARGET),
            attributes=[
                ElementAttribute(field="Company name"),
                YAttribute(field="Amount"),
                TargetAttribute(field="Target"),
            ],
            data=EXAMPLE_DATA_LINE,
        ),
        LineSlide(
            title="Line with target fill area above",
            settings=LineSlideSettings(display=LineSlideDisplay.TARGET, targetFillArea=TargetFillArea.ABOVE),
            attributes=[
                ElementAttribute(field="Company name"),
                YAttribute(field="Amount"),
                TargetAttribute(field="Target"),
            ],
            data=EXAMPLE_DATA_LINE,
        ),
        LineSlide(
            title="Line with target absolute variance",
            settings=LineSlideSettings(display=LineSlideDisplay.TARGET, variance=Variance.ABSOLUTE),
            attributes=[
                ElementAttribute(field="Company name"),
                YAttribute(field="Amount"),
                TargetAttribute(field="Target"),
            ],
            data=EXAMPLE_DATA_LINE,
        ),
        LineSlide(
            title="Line with custom variance",
            settings=LineSlideSettings(variance=Variance.CUSTOM_BAR),
            attributes=[
                ElementAttribute(field="Company name"),
                YAttribute(field="Amount"),
                VarianceAttribute(field="Variance"),
            ],
            data=EXAMPLE_DATA_LINE,
        ),
    ]


def _story_column_slides() -> List[Slide]:
    return [
        HeadingSlide(title="Column examples"),
        ColumnSlide(
            title="Column",
            settings=ColumnSlideSettings(display=ColumnSlideDisplay.BASIC),
            attributes=[
                ElementAttribute(field="Company name"),
                YAttribute(field="Amount"),
            ],
            data=EXAMPLE_DATA_GENERIC,
        ),
        ColumnSlide(
            title="Column with series change",
            settings=ColumnSlideSettings(display=ColumnSlideDisplay.CHANGE),
            attributes=[
                ElementAttribute(field="Company name"),
                YAttribute(field="Amount"),
                SeriesAttribute(field="Quarter", from_="Q1", to="Q2"),
            ],
            data=EXAMPLE_DATA_GENERIC,
        ),
        ColumnSlide(
            title="Column with series target",
            settings=ColumnSlideSettings(display=ColumnSlideDisplay.TARGET),
            attributes=[
                ElementAttribute(field="Company name"),
                YAttribute(field="Amount"),
                SeriesAttribute(field="Quarter", from_="Q1", to="Q2"),
            ],
            data=EXAMPLE_DATA_GENERIC,
        ),
        ColumnSlide(
            title="Column with series absolute variance",
            settings=ColumnSlideSettings(display=ColumnSlideDisplay.CHANGE, variance=Variance.ABSOLUTE),
            attributes=[
                ElementAttribute(field="Company name"),
                YAttribute(field="Amount"),
                SeriesAttribute(field="Quarter", from_="Q1", to="Q2"),
            ],
            data=EXAMPLE_DATA_GENERIC,
        ),
    ]


def _story_bar_slides() -> List[Slide]:
    return [
        HeadingSlide(title="Bar examples"),
        BarSlide(
            title="Bar",
            settings=BarSlideSettings(display=BarSlideDisplay.BASIC),
            attributes=[
                ElementAttribute(field="Company name"),
                XAttribute(field="Amount"),
            ],
            data=EXAMPLE_DATA_GENERIC,
        ),
        BarSlide(
            title="Bar with series change",
            settings=BarSlideSettings(display=BarSlideDisplay.CHANGE),
            attributes=[
                ElementAttribute(field="Company name"),
                XAttribute(field="Amount"),
                SeriesAttribute(field="Quarter", from_="Q1", to="Q2"),
            ],
            data=EXAMPLE_DATA_GENERIC,
        ),
        BarSlide(
            title="Bar with series target",
            settings=BarSlideSettings(display=BarSlideDisplay.TARGET),
            attributes=[
                ElementAttribute(field="Company name"),
                XAttribute(field="Amount"),
                SeriesAttribute(field="Quarter", from_="Q1", to="Q2"),
            ],
            data=EXAMPLE_DATA_GENERIC,
        ),
        BarSlide(
            title="Bar with series absolute variance",
            settings=BarSlideSettings(display=BarSlideDisplay.CHANGE, variance=Variance.ABSOLUTE),
            attributes=[
                ElementAttribute(field="Company name"),
                XAttribute(field="Amount"),
                SeriesAttribute(field="Quarter", from_="Q1", to="Q2"),
            ],
            data=EXAMPLE_DATA_GENERIC,
        ),
    ]


EXAMPLE_DATA_GENERIC = [
    {"Company name": "Blockers Mueller Inc.", "Amount": "520490.0", "Quarter": "Q2"},
    {"Company name": "Blockers Mueller Inc.", "Amount": "420490.0", "Quarter": "Q1"},
    {"Company name": "Comic Prime Inc.", "Amount": "447273.0", "Quarter": "Q2"},
    {"Company name": "Comic Prime Inc.", "Amount": "347273.0", "Quarter": "Q1"},
    {"Company name": "Cupar Times Inc.", "Amount": "345540.0", "Quarter": "Q2"},
    {"Company name": "Cupar Times Inc.", "Amount": "245540.0", "Quarter": "Q1"},
    {"Company name": "Bryson Wirral Ltd.", "Amount": "185344.0", "Quarter": "Q2"},
    {"Company name": "Necessities Caunt Co.", "Amount": "133427.0", "Quarter": "Q2"},
    {"Company name": "Maintenence Florence Co.", "Amount": "129326.0", "Quarter": "Q2"},
    {"Company name": "Herbert Geotechnical Co.", "Amount": "118896.0", "Quarter": "Q2"},
    {"Company name": "Currie Hickinbottom Ltd.", "Amount": "77750.0", "Quarter": "Q2"},
    {"Company name": "Pickering Precision Co.", "Amount": "55664.0", "Quarter": "Q2"},
    {"Company name": "Bunting Letterbox Ltd.", "Amount": "38667.0", "Quarter": "Q2"},
    {"Company name": "Stratos Pack Inc.", "Amount": "25650.0", "Quarter": "Q2"},
    {"Company name": "Wisbech Wallpaper Ltd.", "Amount": "24898.0", "Quarter": "Q2"},
    {"Company name": "Olga Biosystems Ltd.", "Amount": "18690.0", "Quarter": "Q2"},
    {"Company name": "Kingdon Denton Inc.", "Amount": "15352.0", "Quarter": "Q2"},
    {"Company name": "Sants Caledonia Inc.", "Amount": "13530.0", "Quarter": "Q2"},
    {"Company name": "Norie Wilts Co.", "Amount": "10542.0", "Quarter": "Q2"},
    {"Company name": "Suzanna Batt Inc.", "Amount": "2778.0", "Quarter": "Q2"},
    {"Company name": "Runner Driven Ltd.", "Amount": "2516.0", "Quarter": "Q2"},
    {"Company name": "Jenson Simon Inc.", "Amount": "0.0", "Quarter": "Q2"},
    {"Company name": "Hussain Subscription Co.", "Amount": "0.0", "Quarter": "Q2"},
    {"Company name": "Diane Grous Ltd.", "Amount": "0.0", "Quarter": "Q2"},
]

EXAMPLE_DATA_LINE = [
    {"Company name": "A", "Amount": "100", "Target": "100", "Variance": "100"},
    {"Company name": "B", "Amount": "200", "Target": "100", "Variance": "100"},
    {"Company name": "C", "Amount": "100", "Target": "300", "Variance": "100"},
    {"Company name": "D", "Amount": "50", "Target": "100", "Variance": "500"},
    {"Company name": "E", "Amount": "-100", "Target": "50", "Variance": "100"},
    {"Company name": "F", "Amount": "-500", "Target": "100", "Variance": "-100"},
]

EXAMPLE_DATA_BUBBLE = [
    {
        "Company name": "Blockers Mueller Inc.",
        "Amount": "520490.0",
        "Employees": 99,
        "Size by metric": 100,
        "Quarter": "Q2",
    },
    {
        "Company name": "Blockers Mueller Inc.",
        "Amount": "320490.0",
        "Employees": 59,
        "Size by metric": 10,
        "Quarter": "Q1",
    },
    {"Company name": "Comic Prime Inc.", "Amount": "447273.0", "Employees": 20, "Size by metric": 20, "Quarter": "Q2"},
    {"Company name": "Comic Prime Inc.", "Amount": "147273.0", "Employees": 90, "Size by metric": 20, "Quarter": "Q1"},
    {"Company name": "Cupar Times Inc.", "Amount": "345540.0", "Employees": 1, "Size by metric": 200, "Quarter": "Q2"},
    {
        "Company name": "Bryson Wirral Ltd.",
        "Amount": "185344.0",
        "Employees": 100,
        "Size by metric": 150,
        "Quarter": "Q2",
    },
    {
        "Company name": "Necessities Caunt Co.",
        "Amount": "133427.0",
        "Employees": 200,
        "Size by metric": 100,
        "Quarter": "Q2",
    },
]


EXAMPLE_DATA_SCATTER = [
    {"Company name": "Blockers Mueller Inc.", "Amount": "520490.0", "Employees": 99, "Quarter": "Q2"},
    {"Company name": "Blockers Mueller Inc.", "Amount": "420490.0", "Employees": 40, "Quarter": "Q1"},
    {"Company name": "Comic Prime Inc.", "Amount": "447273.0", "Employees": 20, "Quarter": "Q2"},
    {"Company name": "Cupar Times Inc.", "Amount": "345540.0", "Employees": 1, "Quarter": "Q2"},
    {"Company name": "Bryson Wirral Ltd.", "Amount": "185344.0", "Employees": 100, "Quarter": "Q2"},
    {"Company name": "Necessities Caunt Co.", "Amount": "133427.0", "Employees": 200, "Quarter": "Q2"},
]
